"""Data Sets."""

from canonical_sets.data.adult import Adult
from canonical_sets.data.base import DataSet
from canonical_sets.data.compas import Compas

__all__ = [
    "Adult",
    "Compas",
    "DataSet",
]
